/*

Copyright (c) 2003, CRunTiny.org
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the names "CRunTiny" and "CRunTiny.org" nor the names of its 
contributors must not be used to endorse or promote products derived from 
this software without prior written permission. For written permission, please 
contact cruntiny@cruntiny.org.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <crtcore.h>
#include <exception>
#include "compiler.h"
#include "eh_guard_descriptor.h"
#include "eh_context.h"
#include "thread_data.h"



namespace __Compiler
{
	namespace EH
	{
		
		static
		void FindHandlerForForeignException(ExceptionContext & context, 
											GuardDescriptor & guard, 
											BasicRegistrationNode * pMarkerRN)
		{
			if (context.ExceptionCode() == STATUS_BREAKPOINT)
				return;

			const int curState = guard.CurrentState();
			
			unsigned int curTry=0, end=0;

			for(const TryBlockMapEntry * pEntry = guard.GetRangeOfTrysToCheck(&curTry,&end); 
				curTry < end;
				++curTry, ++pEntry)
			{
				if (curState < pEntry->tryLow || curState > pEntry->tryHigh)
					continue;
				const HandlerType * const pCatch = &(pEntry->pHandlerArray[pEntry->nCatches - 1]);
				if (!pCatch->pType || !pCatch->pType->name[0])
					guard.Catch(context, pCatch, 0, pEntry, pMarkerRN);
			}
		}



		static
		void FindHandler(ExceptionContext context, 
						 GuardDescriptor & guard,
						 BasicRegistrationNode * pMarkerRN)
		{
			const int curState = guard.CurrentState();

			unsigned int curTry=0, end=0;

			for(const TryBlockMapEntry * pEntry = guard.GetRangeOfTrysToCheck(&curTry, &end); 
				curTry < end;
				++curTry, ++pEntry)
			{
				if (curState < pEntry->tryLow || curState > pEntry->tryHigh)
					continue;

				const HandlerType * pCatch = pEntry->pHandlerArray;

				for(int catches = pEntry->nCatches; catches > 0; --catches,++pCatch)
				{
					const CatchableType * pCatchable = 
						context.ExceptionObject().FindMatch(pCatch);

					if (pCatchable)
					{
						guard.Catch(context,pCatch,pCatchable,pEntry,pMarkerRN);
						break;
					}
				}
			}

			
		}

		EXCEPTION_DISPOSITION CallHandler(HandlerPtr pfn,
										  ExceptionContext & context,
										  const GuardDescriptor & guard,
										  BasicRegistrationNode * pMarkerRN, 
										  bool recursive)
		{
			return pfn(context.m_pRecord,
					   guard.m_pRN,
					   context.m_pContext,
					   context.m_pDispatcherContext,
					   guard.m_pFuncInfo,
					   guard.m_CatchDepth,
					   pMarkerRN,
					   recursive ? 1 : 0);
		}

		
		EXCEPTION_DISPOSITION CxxFrameHandler(ExceptionContext & context, 
											  GuardDescriptor & guard,
											  BasicRegistrationNode * pMarkerRN, 
											  bool recursive)
		{
			if (!guard.IsValid())
				inconsistency();
			const DWORD unwinding_mask = SEH::EH_UNKNOWN3 | 
										 SEH::EH_UNKNOWN2 | 
										 SEH::EH_EXIT_UNWIND | 
										 SEH::EH_UNWINDING;

			if ( context.ExceptionFlags() & unwinding_mask )
			{
				if (guard.MaxState() != 0 && guard.CatchDepth() == 0)
				{
					guard.FrameUnwindToState(-1);
				}
				return ExceptionContinueSearch;
			}

			if (guard.MayBeUnwound())
			{
				HandlerPtr pfn = context.CustomHandler();
				if (pfn)
				{
					return CallHandler(pfn, context, guard, pMarkerRN, recursive);
				}
			}

			const int curState = guard.CurrentState();
			if (curState < -1 || curState >= guard.MaxState())
				inconsistency();

			if (!context.IsCppException())
			{
				if (!recursive)
					std::terminate();
				FindHandlerForForeignException(context, guard, pMarkerRN);
			}
			else
			{
				ExceptionContext ValidContext = context;

				if (!context.ExceptionObject().IsValid())
				{
					ValidContext = ThreadData::GetInstance()->CurrentException();

					if (!ValidContext.IsValid())
						return ExceptionContinueSearch;
					
					if (ValidContext.IsCppException() && 
						!ValidContext.ExceptionObject().IsValid())
						inconsistency();
					
				}
				FindHandler(ValidContext, guard, pMarkerRN);

				if (recursive && ValidContext.IsValid())
					ValidContext.ExceptionObject().Destruct(true);
			}


			return ExceptionContinueSearch;
		}

	}
}

using namespace __Compiler;

extern "C"
__CRUNTINY_EXPORT
EXCEPTION_DISPOSITION __CxxFrameHandler(EH::ExceptionRecord * pExcept, 
										EH::RegistrationNode * pRN, 
										void * pContext, 
										void * pDC)
{
	__asm cld;

	const EH::FuncInfo * pFuncInfo;
	__asm mov pFuncInfo, eax;

	EH::GuardDescriptor guard(pRN, pFuncInfo, 0);
	EH::ExceptionContext context(pExcept, (EH::Context *)pContext, pDC);
	const _EXCEPTION_DISPOSITION result = EH::CxxFrameHandler(context,
															  guard,
															  0,
															  false);
	return result;
}









