/*

Copyright (c) 2003, CRunTiny.org
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the names "CRunTiny" and "CRunTiny.org" nor the names of its 
contributors must not be used to endorse or promote products derived from 
this software without prior written permission. For written permission, please 
contact cruntiny@cruntiny.org.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <crtcore.h>
#include "compiler.h"
#include <typeinfo>
#include "rtti.h"
#include "../algorithms.h"

using namespace __mini_std;

namespace __Compiler
{
	namespace RTTI
	{

		class MatchSIDownCast 
		{
		public:
			MatchSIDownCast(TypeDescriptor * pTargetType):
				m_pTargetType(pTargetType)
			{
			}

			bool operator()(const BaseClassDescriptor * pBase)
			{
				return (*(pBase->pTypeDescriptor) == *m_pTargetType &&
						(pBase->attributes & 1) == 0);
			}
		private:
			TypeDescriptor * m_pTargetType;
		};

		class MatchCrossCast
		{
		public:
			MatchCrossCast(TypeDescriptor * pTargetType):
				m_pTargetType(pTargetType)
			{
			}

			bool operator()(const BaseClassDescriptor * pBase);
		private:
			TypeDescriptor * m_pTargetType;
		};


		class MatchSubBase
		{
		public:
			MatchSubBase(TypeDescriptor * pSrcType,void * pCompleteObject, ptrdiff_t SrcOffset):
				m_pSrcType(pSrcType),
				m_pCompleteObject(pCompleteObject),
				m_SrcOffset(SrcOffset)
			{
			}
			bool operator()(const BaseClassDescriptor * pSubBase);
		private:
			TypeDescriptor * m_pSrcType;
			void * m_pCompleteObject;
			ptrdiff_t m_SrcOffset;
		};

		class MatchMIDownCast
		{
		public:
			MatchMIDownCast(TypeDescriptor * pTargetType,TypeDescriptor * pSrcType,void * pCompleteObject, ptrdiff_t SrcOffset):
				m_pTargetType(pTargetType),
				m_MatchSubBase(pSrcType,pCompleteObject,SrcOffset)
			{
			}

			bool operator()(const BaseClassDescriptor * const * pBasePtr);
		private:
			TypeDescriptor * m_pTargetType;
			MatchSubBase m_MatchSubBase;
		};

		bool MatchCrossCast::operator()(const BaseClassDescriptor * pBase)
		{
			return (*(pBase->pTypeDescriptor) == *m_pTargetType &&
					(pBase->attributes & 1) == 0 &&
					(pBase->attributes & 2) == 0);
		}

		bool MatchSubBase::operator()(const BaseClassDescriptor * pSubBase)
		{
			if (*(pSubBase->pTypeDescriptor) == *m_pSrcType)
			{
				return (AdjustPointer(m_pCompleteObject, pSubBase->where) ==
										(BYTE*)m_pCompleteObject + m_SrcOffset);
			}
			return false;
		}

		bool MatchMIDownCast::operator()(const BaseClassDescriptor * const * pBasePtr)
		{
			typedef const BaseClassDescriptor * const * desc_ptr_ptr;
			typedef const BaseClassDescriptor * desc_ptr;

			desc_ptr const pBase = *pBasePtr;
			if (*(pBase->pTypeDescriptor) == *m_pTargetType)
			{
				desc_ptr_ptr const pSubBegin = pBasePtr + 1;
				desc_ptr_ptr const pSubEnd = pSubBegin + pBase->numContainedBases;

				desc_ptr_ptr const pSubBasePtr = find_if(pSubBegin, pSubEnd, m_MatchSubBase);
				
				return (pSubBasePtr != pSubEnd);
			}
			return false;
		}
		

		void * FindCompleteObject(void * pObject, CompleteObjectLocator * pCompleteLocator)
		{
			BYTE * pCompleteObject = (BYTE *)pObject - pCompleteLocator->offset;
			if (pCompleteLocator->cdOffset != 0) 
				pCompleteObject += *(DWORD*)((BYTE *)pObject - pCompleteLocator->cdOffset);
			return pCompleteObject;
		}

		const BaseClassDescriptor * 
			FindSITargetTypeInstance(void * /*pCompleteObject*/, 
								     const CompleteObjectLocator * pCompleteLocator, 
									 TypeDescriptor * /*pSrcType*/, 
									 ptrdiff_t /*SrcOffset*/, 
									 TypeDescriptor * pTargetType)
		{
			typedef const BaseClassDescriptor * const * desc_ptr_ptr;
			typedef const BaseClassDescriptor * desc_ptr;

			desc_ptr_ptr const pBegin = 
				pCompleteLocator->pClassDescriptor->pBaseClassArray->arrayOfBaseClassDescriptors;
			desc_ptr_ptr const pEnd = 
				pBegin + pCompleteLocator->pClassDescriptor->numBaseClasses;

			desc_ptr_ptr const pBasePtr = find_if(pBegin, pEnd, MatchSIDownCast(pTargetType));
			return (pBasePtr != pEnd ? *pBasePtr : 0);
		}

		const BaseClassDescriptor * 
			FindMITargetTypeInstance(void * pCompleteObject, 
								     const CompleteObjectLocator * pCompleteLocator, 
									 TypeDescriptor * pSrcType, 
									 ptrdiff_t SrcOffset, 
									 TypeDescriptor * pTargetType)
		{
			typedef const BaseClassDescriptor * const * desc_ptr_ptr;
			typedef const BaseClassDescriptor * desc_ptr;

			desc_ptr_ptr const pBegin = 
				pCompleteLocator->pClassDescriptor->pBaseClassArray->arrayOfBaseClassDescriptors;
			desc_ptr_ptr const pEnd =
				pBegin + pCompleteLocator->pClassDescriptor->numBaseClasses;

			desc_ptr_ptr pBasePtr = 
				find_if_ptr(pBegin, pEnd, MatchMIDownCast(pTargetType,pSrcType,pCompleteObject,SrcOffset));
			if (pBasePtr != pEnd)
				return *pBasePtr;

			pBasePtr = find_if(pBegin, pEnd, MatchCrossCast(pTargetType));
			return (pBasePtr != pEnd ? *pBasePtr : 0);
		}


		const BaseClassDescriptor * 
			FindVITargetTypeInstance(void * pCompleteObject, 
								     const CompleteObjectLocator * pCompleteLocator, 
									 TypeDescriptor * pSrcType, 
									 ptrdiff_t SrcOffset, 
									 TypeDescriptor * pTargetType)
		{
			typedef const BaseClassDescriptor * const * desc_ptr_ptr;
			typedef const BaseClassDescriptor * desc_ptr;


			desc_ptr_ptr const pBegin = 
				pCompleteLocator->pClassDescriptor->pBaseClassArray->arrayOfBaseClassDescriptors;
			desc_ptr_ptr const pEnd =
				pBegin + pCompleteLocator->pClassDescriptor->numBaseClasses;

			desc_ptr pResult = 0;
			{//fuck vc6
			for(desc_ptr_ptr pBasePtr = pBegin; pBasePtr != pEnd; ++pBasePtr)
			{
				desc_ptr const pBase = *pBasePtr;
				if (*(pBase->pTypeDescriptor) == *pTargetType)
				{
					desc_ptr_ptr const pSubBegin = pBasePtr + 1;
					desc_ptr_ptr const pSubEnd = pSubBegin + pBase->numContainedBases;

					for(desc_ptr_ptr pSubBasePtr = pSubBegin; pSubBasePtr != pSubEnd; ++pSubBasePtr)
					{
						desc_ptr const pSubBase = *pSubBasePtr;

						if (*(pSubBase->pTypeDescriptor) == *pSrcType)
						{
							if (AdjustPointer(pCompleteObject,pSubBase->where) ==
								(BYTE*)pCompleteObject + SrcOffset)
							{
								if (pResult)
								{
									if (AdjustPointer(pCompleteObject, pResult->where) !=
										AdjustPointer(pCompleteObject, pBase->where))
										return 0;
								}
								pResult = pBase;
							}
						}
					}
				}
			}
			}//end fuck vc6

			if (pResult)
				return pResult;

			desc_ptr_ptr pBasePtr = find_if(pBegin, pEnd, MatchCrossCast(pTargetType));
			return (pBasePtr != pEnd ? *pBasePtr : 0);
		}
	}
}

using namespace __Compiler::RTTI;

extern "C"
void * __RTDynamicCast(void * inptr, 
					   long VfDelta, 
					   void * SrcType, 
					   void * TargetType, 
					   int isReference)
{
	typedef const BaseClassDescriptor * 
			(*FindTargetTypeInstance)(void * pCompleteObject, 
								      const CompleteObjectLocator * pCompleteLocator, 
									  TypeDescriptor * pSrcType, 
									  ptrdiff_t SrcOffset, 
									  TypeDescriptor * pTargetType);

	if (inptr == 0)
		return 0;
	CompleteObjectLocator * pCompleteLocator = (*(CompleteObjectLocator***)inptr)[-1];
	void * pCompleteObject = FindCompleteObject(inptr, pCompleteLocator);

	inptr = (BYTE*)inptr - VfDelta;
	ptrdiff_t inptr_delta = (BYTE*)inptr - (BYTE*)pCompleteObject;
	const BaseClassDescriptor * pBaseClass = 0;
	FindTargetTypeInstance Find = 0;
	
	if ( (pCompleteLocator->pClassDescriptor->attributes & 1) == 0)
		Find = FindSITargetTypeInstance;
	else if ( (pCompleteLocator->pClassDescriptor->attributes & 2) == 0 )
		Find = FindMITargetTypeInstance;
	else
		Find = FindVITargetTypeInstance;

	pBaseClass = Find(pCompleteObject, 
					  pCompleteLocator, 
					  (TypeDescriptor *)SrcType, 
					  inptr_delta, 
					  (TypeDescriptor *)TargetType);

	if (pBaseClass == 0)
	{
		if (isReference)
			throw std::bad_cast("Bad dynamic_cast!");
		return 0;
	}

	return AdjustPointer(pCompleteObject, pBaseClass->where);	
}








