/*

Copyright (c) 2003, CRunTiny.org
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the names "CRunTiny" and "CRunTiny.org" nor the names of its 
contributors must not be used to endorse or promote products derived from 
this software without prior written permission. For written permission, please 
contact cruntiny@cruntiny.org.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <crtcore.h>
#include "compiler.h"
#include "exception_handling.h"
#include "thread_data.h"
#include "call.h"

namespace __Compiler
{
	namespace EH
	{

		const DWORD CPP_EXCEPTION	= 0xE06D7363;
		const DWORD MS_MAGIC		= 0x19930520;
		const DWORD MS_MAGIC_MASK	= 0x1FFFFFFF;


		void ThrownObject::Destruct(bool fThrowNotAllowed)
		{
			if (m_pObjectInfo)
			{
				__try
				{
					CallMemberFunction(m_pObject, m_pObjectInfo->pmfnUnwind);
				}
				__except(fThrowNotAllowed ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
				{
					std::terminate();
				}
			}
		}

		void ThrownObject::Copy(void ** pCatchBuffer, 
								const HandlerType * pCatch, 
								const CatchableType * pConv)
		{
			if (pCatch->adjectives & 8)    //catch by pointer ??
			{
				/*if (!ValidateRead(pObject(), 1) ||
					!ValidateWrite(pCatchBuffer, 1))
					inconsistency();*/
				*pCatchBuffer = m_pObject;
				*pCatchBuffer = RTTI::AdjustPointer(*pCatchBuffer,pConv->thisDisplacement);
			}
			else if (pConv->properties & 1)  //pointer -?
			{
				/*if (!ValidateRead(pObject(), 1) ||
						!ValidateWrite(pCatchBuffer,1))
						inconsistency(); */
				memmove(pCatchBuffer, m_pObject, pConv->sizeOrOffset);
				if (pConv->sizeOrOffset == 4 && *pCatchBuffer != 0)
					*pCatchBuffer = RTTI::AdjustPointer(*pCatchBuffer,pConv->thisDisplacement);
			}
			else if (pConv->copyFunction == 0)
			{
				/*if (!ValidateRead(pObject(), 1) ||
					!ValidateWrite(pCatchBuffer, 1))
					inconsistency(); */
				memmove(pCatchBuffer, 
						RTTI::AdjustPointer(m_pObject,pConv->thisDisplacement),
						pConv->sizeOrOffset); 
			}
			else
			{
				/*if (!ValidateRead(pObject(), 1) ||
						!ValidateWrite(pCatchBuffer, 1) ||
						!ValidateExecute(pConv->copyFunction))
						inconsistency(); */
				if (pConv->properties & 4) //array member - ?
				{
					CallMemberFunction(pCatchBuffer, 
									   pConv->copyFunction, 
									   RTTI::AdjustPointer(m_pObject, pConv->thisDisplacement), 
									   1);
				}
				else
				{
					CallMemberFunction(pCatchBuffer, 
									   pConv->copyFunction, 
									   RTTI::AdjustPointer(m_pObject, pConv->thisDisplacement));
				}
			}
		}

		const CatchableType * ThrownObject::FindMatch(const HandlerType * pCatch)
		{
			const CatchableTypeArray * const pCatchableTypeArray =
				m_pObjectInfo->pCatchableTypeArray;

			const CatchableType * const * ppCatchable = 
				pCatchableTypeArray->arrayOfCatchableTypes;
			
			for(int catchables = pCatchableTypeArray->nCatchableTypes ; 
				catchables > 0; 
				--catchables, ++ppCatchable)
			{
				const CatchableType * pCatchable = *ppCatchable;
				if (TypeMatch(pCatch, pCatchable, m_pObjectInfo))
				{
					return pCatchable;
				}
			}

			return 0;
		}

		bool ThrownObject::TypeMatch(const HandlerType * pCatch, 
									 const CatchableType * pCatchable, 
									 const ThrowInfo * pThrow)
		{
			//if this is a catch(...) it matches everything
			if (pCatch->pType == 0 ||
				pCatch->pType->name[1] == 0)
				return true;
			
			if (pCatch->pType != pCatchable->pType &&
				strcmp(pCatch->pType->name, pCatchable->pType->name) != 0)
			{
				return false;
			}

			if ((pCatchable->properties & 2) != 0 &&
				(pCatch->adjectives & 8) == 0)
			{
				return false;
			}

			if ((pThrow->attributes & 1) != 0 &&
				(pCatch->adjectives & 1) == 0)
			{
				return false;
			}

			if ((pThrow->attributes & 2) != 0 &&
				(pCatch->adjectives & 2) == 0)
			{
				return false;
			}

			return true;
		}
		
		

		ExceptionRecord::ExceptionRecord(ThrownObject & object, void * pExceptionAddress)
		{
			ExceptionCode = CPP_EXCEPTION;
			ExceptionFlags = EXCEPTION_NONCONTINUABLE;
			_EXCEPTION_RECORD::ExceptionRecord = 0;
			ExceptionAddress = pExceptionAddress;
			NumberParameters = 3;
			ExceptionInformation[0] = MS_MAGIC;
			Object() = object;
		}

		void ExceptionRecord::CopyObject(RegistrationNode * pRN, 
										 const HandlerType * pCatch, 
										 const CatchableType * pConv)
		{
			if (pCatch->pType != 0 && 
				pCatch->pType->name[0] != 0 &&
				( pCatch->dispCatchObj != 0 || 
				  (pCatch->adjectives & 0x80000000) != 0 ))
			{
				void ** pCatchBuffer = (void **)((pCatch->adjectives & 0x80000000) == 0 ?
											     (BYTE*)(pRN->GetStackFrame()) + pCatch->dispCatchObj :
											     (BYTE*)pRN);
				__try
				{
					Object().Copy(pCatchBuffer, pCatch, pConv);
				}
				__except(EXCEPTION_EXECUTE_HANDLER)
				{
					std::terminate();
				}
			}
		}


		__declspec(noreturn)
		void inconsistency()
		{
			std::terminate();
		}
	}

} // namespace __Compiler









