/*

Copyright (c) 2003, CRunTiny.org
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the names "CRunTiny" and "CRunTiny.org" nor the names of its 
contributors must not be used to endorse or promote products derived from 
this software without prior written permission. For written permission, please 
contact cruntiny@cruntiny.org.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <crtcore.h>
#include "call.h"
#include "seh.h"

namespace __Compiler
{

	namespace SEH
	{	
		const DWORD EH_NONCONTINUABLE	= 0x01;
		const DWORD EH_UNWINDING		= 0x02;
		const DWORD EH_EXIT_UNWIND		= 0x04;
		const DWORD EH_STACK_INVALID	= 0x08;
		const DWORD EH_NESTED_CALL		= 0x10;
		const DWORD EH_UNKNOWN1			= 0x20;
		const DWORD EH_UNKNOWN2			= 0x20;
		const DWORD EH_UNKNOWN3			= 0x40;

		

		EXCEPTION_DISPOSITION UnwindHandler(ExceptionRecord * pExcept,
											ExceptionFrame * pFrame,
											Context * /*pContext*/,
											void * pDC)
		{
			if (!pExcept->IsUnwinding())
				return ExceptionContinueSearch;
			*(void**)pDC = pFrame->Next();
			return ExceptionCollidedUnwind;
		}

		

		void LocalUnwind(ExceptionFrame * pFrame, int tryLevel)
		{
			UnwindFrame local_frame(UnwindHandler, -2, pFrame);
			local_frame.SetCurrent();

			while (pFrame->TryLevel() != -1 &&
				   pFrame->TryLevel() != tryLevel)
			{
				
				const ScopeTableEntry & entry = pFrame->CurrentEntry();
				local_frame.SetTryLevel(entry.previousTryLevel);
				pFrame->SetTryLevel(entry.previousTryLevel);

				if (entry.previousTryLevel != 0)
					continue;

				entry.lpfnHandler();
			}
			
			local_frame.Remove();

		}


		EXCEPTION_DISPOSITION ExceptionHandler(ExceptionRecord * pExcept,
											   ExceptionFrame * pFrame,
											   Context * pContext,
											   void * /*pDC*/)
		{
			EXCEPTION_POINTERS ex_ptrs;
			ex_ptrs.ContextRecord = pContext;
			ex_ptrs.ExceptionRecord = pExcept;

			pFrame->SetExceptionPointers(&ex_ptrs);

			const ScopeTableEntry * pEntry = 0;

			for (int tryLevel = pFrame->TryLevel(); 
				 tryLevel != -1; 
				 tryLevel = pEntry->previousTryLevel)
			{
				pEntry = &pFrame->Entry(tryLevel);

				if (!pEntry->lpfnFilter)
					continue;

				void * const new_ebp = pFrame->EBP();

				const int FilterResult = (int)CallOnFrame0(pEntry->lpfnFilter, new_ebp); 

				if (FilterResult < 0 ) // EXCEPTION_CONTINUE_EXECUTION
					return ExceptionContinueExecution;

				if (FilterResult != EXCEPTION_CONTINUE_SEARCH)
				{
					// If we get here, EXCEPTION_EXECUTE_HANDLER was specified
					__Platform::UnwindStack(pFrame, 0, 0);

					CallOnFrame2(LocalUnwind, new_ebp, pFrame, (void*)(unsigned _W64)tryLevel);

					pFrame->SetTryLevel(pEntry->previousTryLevel);

					CallOnFrame0(pEntry->lpfnHandler, new_ebp);
				}

			}

			return ExceptionContinueSearch;
		}

		EXCEPTION_DISPOSITION CleanupHandler(ExceptionRecord * /*pExcept*/,
											 ExceptionFrame * pFrame,
											 Context * /*pContext*/,
											 void * /*pDC*/)
		{
			void * const new_ebp = pFrame->EBP();
			CallOnFrame2(LocalUnwind, new_ebp, pFrame, (void*)-1);
			return ExceptionContinueSearch;
		}

		BOOL AbnormalTermination()
		{
			const __Platform::ExceptionFrame * pCurrentFrame = 
				__Platform::ExceptionFrame::GetCurrent();
			if (pCurrentFrame->Handler() != UnwindHandler)
				return FALSE;
			const UnwindFrame * pUnwindFrame = (const UnwindFrame *)pCurrentFrame;
			const ExceptionFrame * pOriginalFrame = pUnwindFrame->OriginalFrame();
			if (pUnwindFrame->TryLevel() != pOriginalFrame->TryLevel())
				return FALSE;
			return TRUE;
		}
	}

}

extern "C"
{
	using namespace __Compiler::SEH;

	__CRUNTINY_EXPORT
	EXCEPTION_DISPOSITION __cdecl _except_handler3(ExceptionRecord * pExcept,
												   ExceptionFrame * pFrame,
												   Context * pContext,
												   void * pDC)
	{
		__asm cld;

		if (pExcept->IsUnwinding())
			return CleanupHandler(pExcept, pFrame, pContext, pDC);
		return ExceptionHandler(pExcept, pFrame, pContext, pDC);

	}

}






